#!/usr/bin/env python
import os
import ast
import copy

DEFINITION_FILE_OASS     = "input/gadgetDefinitions-preprocessed.oass"
DEFINITION_FILE_ORIT     = "input/gadgetDefinitions-preprocessed.orit"
DEFINITION_FILE_TXT      = "input/gadgetDefinitions-preprocessed.txt"
DEFINITION_OUTFILE_OASS  = "input/gadgetDefinitions.oass"
DEFINITION_OUTFILE_ORIT  = "input/gadgetDefinitions.orit"
DEFINITION_OUTFILE_TXT   = "input/gadgetDefinitions.txt"


def convertOass():
    
    print "Converting " + DEFINITION_FILE_OASS + " ... ",
    fin = open(DEFINITION_FILE_OASS, 'r')
    lines = fin.readlines()
    fin.close()

    oassDict = {}
    for line in lines:
        l = line.split("=")
        key = l[0]
        seq = l[1].strip()
        oassDict[key] = seq

    color = 'r'

    outBuffer = ""

    assembly = ast.literal_eval(oassDict["Assembly"])

    for bead in assembly:
        bead[0] = bead[0] + color

    oassDict["Assembly"] = str(assembly)

    '''
    OritatamiSystemDef=...
    Assembly=...
    LastBead=...
    '''
    outBuffer = outBuffer + "OritatamiSystemDef="   + oassDict["OritatamiSystemDef"]  + "\n" 
    outBuffer = outBuffer + "Assembly="             + oassDict["Assembly"]             + "\n" 
    outBuffer = outBuffer + "LastBead="             + oassDict["LastBead"] 

    fout = open(DEFINITION_OUTFILE_OASS, "w")
    fout.write(outBuffer)
    fout.close()

    print "Done!"


def convertTxt():
    
    print "Converting " + DEFINITION_FILE_TXT + " ... ",
    fin = open(DEFINITION_FILE_TXT, 'r')
    lines = fin.readlines()
    fin.close()

    gadgetDict = {}
    for line in lines:
        l = line.split("=")
        key = l[0]
        seq = l[1].strip()
        gadgetDict[key] = seq

    colors = ['r','g','b']

    outBuffer = ""

    for color in colors:
        for key in gadgetDict:
            line = key + color + "="
            first = True
            for bead in gadgetDict[key].replace(" ","").replace("'", "").split(","):
                 if first:
                    line = line + "'" + bead + color + "'"
                    first = False
                 else:
                    line = line +  ",'" + bead + color + "'"

            outBuffer = outBuffer + line + "\n"

    fout = open(DEFINITION_OUTFILE_TXT, "w")
    fout.write(outBuffer)
    fout.close()
    print "Done!"






'''
gadget schemes
1='1','2','3','4','5','6','x','7','x','12','8','x','13','9','x','14','10','x','11'
2='15','16','17','18','19','20','x','21','x','28','22','29','27','x','23','24','x','25','26'
3='30','31','32','33','34','35','44','36','45','37','x','38','39','40','46','x','41','42','43'
4='47','48','49','50','51','52','x','53','78','54','55','56','57','76','58','59','x','60','61'
5='62','63','64','65','66','67','x','68','79','69','70','80','71','72','x','73','74','77','75'
'''
def convertOrit():
    
    print "Converting " + DEFINITION_FILE_ORIT + " ... ",


    fin = open(DEFINITION_FILE_TXT, 'r')
    lines = fin.readlines()
    fin.close()

    gadgetDict = {}
    allBeadTypes = []
    for line in lines:
        l = line.split("=")
        key = l[0]
        seq = l[1].strip()
        gadgetDict[key] = seq
        for bd in seq.replace(" ","").replace("'","").split(","):
            allBeadTypes.append(bd)

    fin = open(DEFINITION_FILE_ORIT, 'r')
    lines = fin.readlines()
    fin.close()

    oritDict = {}
    for line in lines:
        l = line.split("=")
        key = l[0]
        value = l[1].strip()
        oritDict[key] = value


    beadTypes = ast.literal_eval(oritDict["BeadTypes"])

    colors = ['r','g','b']
    colorVals = {}

    colorVals['r'] = "(255,0,0)"
    colorVals['g'] = "(0,255,0)"
    colorVals['b'] = "(0,0,255)"

    newBeadTypes = {}



    # first second and third beads in a gadget need to be able to bind to these.
    bindsToFirst  = ['60', '121', '182', '243', '254']
    bindsToSecondAndThird = ['59', '120', '181', '242', '253']

    first  = ['1', '62', '123', '184']
    secondOrThird = ['2', '3', '63', '64', '124', '125', '185', '186']

    # gadget 5 is a special case
    initGadget5 = ['245', '246', '247'] # the first 3 beads of gadget 5
    bindsToFirstGadget5  = ['254', '60', '121', '182', '243'] # these bind to the first bead of gadget 5
    bindsToSecondAndThirdGadget5 = ['304', '10', '89', '171', '235'] # these bind to the 2nd and 3rd bead of gadget 5


    for colorIdx in range(len(colors)): # the index comes in handy later
        color = colors[colorIdx]
        tmpBeadTypes = copy.deepcopy(beadTypes)
 
                        
        for beadType in tmpBeadTypes:
            for key in gadgetDict:
                
                if str(beadType[0]) in gadgetDict[key].replace(" ", "").replace("'", "").split(","):

                    origBead = beadType[0]

                    beadType[0] = beadType[0] + color
                    beadType[1] = colorVals[color]
    
                    tmpBonds = []
                    for i in range(len(beadType[2])):
                        if beadType[2][i] in gadgetDict[key].replace(" ", "").replace("'", "").split(","):
                            tmpBonds.append(beadType[2][i] + color)                        
                    beadType[2] = tmpBonds

                    # the following bonds serve as connectors between gadgets
                    if origBead in first:
                        for bd in bindsToFirst:
                            beadType[2].append(bd + colors[(colorIdx+1)%3])
                            beadType[2].append(bd + colors[(colorIdx+2)%3])
                    elif origBead in secondOrThird:
                        for bd in bindsToSecondAndThird:
                            beadType[2].append(bd + colors[(colorIdx+1)%3])
                            beadType[2].append(bd + colors[(colorIdx+2)%3])

                    # gadget 5 requires special cases since it is "turned around"
                    elif origBead == initGadget5[0]:
                        for bd in bindsToFirstGadget5:
                            beadType[2].append(bd + colors[(colorIdx+1)%3])
                            beadType[2].append(bd + colors[(colorIdx+2)%3])
                    elif origBead == initGadget5[1] or origBead == initGadget5[2]:
                        for bd in bindsToSecondAndThirdGadget5:
                            beadType[2].append(bd + colors[(colorIdx+1)%3])
                            beadType[2].append(bd + colors[(colorIdx+2)%3])
                                               

        newBeadTypes[color] = list(tmpBeadTypes)

    
    
    
    oritDict["BeadTypes"] = str(newBeadTypes['r'] + newBeadTypes['g'] + newBeadTypes['b'])
    outBuffer = ""

    '''
    BeadTypes=...
    BeadCycle=[]
    DelayLength=3
    Arity=4
    LockWithoutBinding=False
    BindFullFloppySegment=False
    '''

    outBuffer = outBuffer + "BeadTypes="             + oritDict["BeadTypes"]             + "\n" 
    outBuffer = outBuffer + "BeadCycle="             + "[]"                              + "\n" 
    outBuffer = outBuffer + "DelayLength="           + oritDict["DelayLength"]           + "\n" 
    outBuffer = outBuffer + "Arity="                 + oritDict["Arity"]                 + "\n" 
    outBuffer = outBuffer + "LockWithoutBinding="    + oritDict["LockWithoutBinding"]    + "\n" 
    outBuffer = outBuffer + "BindFullFloppySegment=" + oritDict["BindFullFloppySegment"] + "\n" 
    
    fout = open(DEFINITION_OUTFILE_ORIT, "w")
    fout.write(outBuffer)
    fout.close()
    print "Done!"





convertTxt()
convertOrit()
convertOass()
