from ShapeMakerGrid import *
from ShapeMakerNodes import *
from ShapeMakerUtils import *
from SimulatorFramework import SimCanvas
from SimulatorFramework import SimUtils
from LookupTable import *
import wx
import threading
import os
import sys
import random
import math


DEFINITION_FILE_OASS = "input/gadgetDefinitions.oass"
DEFINITION_FILE_ORIT = "input/gadgetDefinitions.orit"
DEFINITION_FILE_TXT  = "input/gadgetDefinitions.txt"

SHOW_GRID_AND_COORDS = 0
SHOW_GRID_ONLY       = 1
SHOW_NO_GRID         = 2


ORIT_MODE_PAN_ZOOM     = 0
ORIT_MODE_DRAW_BEAD    = 1
ORIT_MODE_DRAW_LINK    = 2
ORIT_MODE_DRAW_BOND    = 3

SIMFW_FULLDISPLAY      = 0
#SIMFW_GROWTHDISPLAY    = 1
SIMFW_NODISPLAY        = 1


class ShapeMakerCanvas(SimCanvas.Canvas):
    def __init__(self, parent, idIn, frame, size, app):

        self.dictIdToType = {}
        self.boundsdict = {}
        self.confirmOverArity = True
        self.confirmInvalidBonds = True
        
        self.showGrid = SHOW_GRID_AND_COORDS
        
        self.mouseMode = ORIT_MODE_PAN_ZOOM
        self.drawingModeBeadType = ShapeNodeType

        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("RED", 1)
        self.axesPen = self.ColoredPen("RED", 1)
        self.bondPen = self.ColoredPen("RED", 2)
        self.linkPen = self.ColoredPen("BLACK", 4)
        self.cyclePen = self.ColoredPen("RED", 8)
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)
        
        self.sideLen = TRISIZE
        self.triangleHeight = math.tan(math.radians(60)) * 0.5 * self.sideLen

        #self.gridWidth = GRIDSTARTSIZE
        #self.gridHeight = GRIDSTARTSIZE
        
        self.dragStart   = None
        self.linkStart   = None
        self.linkEnd     = None
        self.posSelected = None
        self.negSelected = None
        self.lastCurrent = None

        SimCanvas.Canvas.__init__(self, parent, idIn, frame, size, app)
        self.SetScrollRate(1,1)
		
    def CreateNewAssembly(self):
        self.assembly = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        self.gridWidth = self.assembly.GetWidth()
        self.gridHeight = self.assembly.GetHeight()
        self.winWidth  = self.sideLen * self.gridWidth  + self.GetHorizOffset(self.gridWidth)
        self.winHeight = self.GetVertOffset(self.gridHeight)

    def DoLoad(self, bAssembly=True, bSystem=False):
        typeString = "Shape definition files (*.opts)|*.opts"
        openFileDialog = wx.FileDialog(self, "Open", "", "", 
                                       typeString,
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):

            self.assembly = ONodeGrid(GRIDSTARTSIZE,GRIDSTARTSIZE)
            self.assembly.LoadOritatamiAssembly(fileName, True)

                #if len(self.assembly.ONodeTypes) > 0:
                    # Randomly set the current drawing bead type to one of the bead types
                    #self.drawingModeBeadType = self.assembly.ONodeTypes[self.assembly.ONodeTypes.keys()[0]]
                #else:
            self.InitializeDrawing()
            self.Refresh()
            self.ScrollToCoords((0,0))

    def DoSaveShapeDef(self, bSaveAs=False):
        fileName = SimUtils.globalApp.systemDefinitionFileFullPath
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save Shape Definition As", "", "", 
                                           "Shape definition files (*.opts)|*.opts",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                SimUtils.globalApp.ShowErrorDialog("You must provide a file name to save the shape definition to.")
                return False
            fileName = saveFileDialog.GetPath()
            SimUtils.globalApp.systemDefinitionFile = fileName
            saveFileDialog.Destroy()

        SimUtils.globalApp.systemDefinitionFileFullPath = fileName
        if os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            btStr = "["
            for bt in self.assembly.ONodeTypes:
                if len(btStr) > 1:
                    btStr = btStr + ", "
                btStr = btStr + self.assembly.ONodeTypes[bt].Output()
            btStr = btStr + "]"
            btLine = "BeadTypes=" + btStr
            #print("btLine: " + btLine)
            assemblyString = "Assembly="
            #lastBeadString = "LastBead=" + str(self.assembly.lastBeadLoc)
            assemblyList = self.GetAssemblyList(bOnlySeed=False)
            if len(assemblyList) == 0:
                SimUtils.globalApp.ShowErrorDialog("The shape definition is empty.  It must contain at least one point.")
                return False
            assemblyString = assemblyString + str(assemblyList)

            wFile = open(fileName, "w")
            #wFile.write("OritatamiSystemDef=shape-definition.orit\n")
            wFile.write(assemblyString + "\n")
            #wFile.write(lastBeadString + "\n")
            wFile.close()
            SimUtils.globalApp.OutputMessage("Saved shape definition to file '" + fileName + "'")
            return True
        except:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: Unable to write shape definition file: '" + fileName + "'\n")
            return False

    def DoSetSeed(self):
        dlg = wx.MessageDialog(self, "Are you sure you want to set the current assembly as the seed assembly?", caption="Confirm seed setting")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            assemblyList = self.GetAssemblyList()
            if len(assemblyList) == 0:
                SimUtils.globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
                return False
            self.assembly.seed = assemblyList
            self.assembly.lastSeedLoc = self.assembly.lastBeadLoc
        else:
            dlg.Destroy()

    #  Format for assembly file:
    #  [[<bead location>], [<bead location>], ...]
    #  Format for bead location:
    #  ["<type name>", (<coords>), (<parent coords>), [<list of coords bonded to>]]
    def DoSaveAssembly(self, bOnlySeed=False, bSaveAs=False):
        aStr = "assembly"
        AStr = "Assembly"
        fileName = SimUtils.globalApp.assemblyDefinitionFile
        if bOnlySeed == True:
            aStr = "seed"
            AStr = "Seed"
            fileName = SimUtils.globalApp.seedDefinitionFile
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save " + AStr + " As", "", "", 
                                           "Oritatami " + aStr + " files (*.oass)|*.oass",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                return False
            fileName = saveFileDialog.GetPath()
            saveFileDialog.Destroy()

        if SimUtils.globalApp.systemDefinitionFile is None or len(SimUtils.globalApp.systemDefinitionFile) < 1:
            SimUtils.globalApp.ShowErrorDialog("\nERROR: In order to save an assembly, you must first have a valid oritatami system definition file associated with it. Please save your system definition first.\n")
            return False

        systemDefString = "OritatamiSystemDef=" + SimUtils.globalApp.systemDefinitionFile
        assemblyString = "Assembly="
        if bOnlySeed == True:
            lastBeadString = "LastBead=" + str(self.assembly.lastSeedLoc)
        else:
            lastBeadString = "LastBead=" + str(self.assembly.lastBeadLoc)
        assemblyList = self.GetAssemblyList(bOnlySeed)
        if len(assemblyList) == 0:
            SimUtils.globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
            return False
        assemblyString = assemblyString + str(assemblyList)
        
        origFileName = fileName
        if not fileName is None:
            if len(fileName) < 1:
                return False
            elif os.path.isabs(fileName) == False:
                fileName = "./" + fileName
                fileName = os.path.abspath(fileName)

            try:
                wFile = open(fileName, "w")
                wFile.write(systemDefString + "\n")
                wFile.write(assemblyString + "\n")
                wFile.write(lastBeadString)
                wFile.close()
                if bOnlySeed == True:
                    SimUtils.globalApp.seedDefinitionFile = origFileName
                else:
                    SimUtils.globalApp.assemblyDefinitionFile = origFileName
                SimUtils.globalApp.OutputMessage("Saved oritatami " + aStr + " to file '" + fileName + "'")
                return True
            except:
                SimUtils.globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami " + aStr + " to file: '" + fileName + "'\n")
                return False

    def GetAssemblyList(self, bOnlySeed=False):
        assemblyList = []
        if bOnlySeed == True:
            assemblyList = self.assembly.seed
        else:
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                for j in range(self.assembly.minY, self.assembly.maxY+1):
                    locList = self.GetLocOutputList((i,j))
                    if not locList is None:
                        assemblyList.append(locList)
        return assemblyList

    def GetLocOutputList(self, loc):
        onode = self.assembly.GetLoc(loc)
        if onode is not None:
            #self.assembly.ONodeTypes[onode.type]
            locList = []
            locList.append(onode.type)
            locList.append(loc)
            if onode.parent is None:
                locList.append(None)
            else:
                locList.append(onode.parent.coords)
            locList.append(onode._bonds)
            return locList
        else:
            return None
        
    def DoSaveAll(self):
        
        if self.DoSaveShapeDef(bSaveAs=False) == True:
            #if self.DoSaveAssembly(bOnlySeed=True, bSaveAs=False) == True:
            return True
        return False

    def DoNew(self):
        
        self.assembly = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)

        self.SetRandomSeed(self.randSeed)
        SimUtils.globalApp.seedDefinitionFile = None
        SimUtils.globalApp.assemblyDefinitionFile = None
        SimUtils.globalApp.systemDefinitionFile = None
        SimUtils.globalApp.systemDefinitionFileFullPath = None
        
        self.drawingModeBeadType = ShapeNodeType
        self.InitializeDrawing()
        self.drawingModeBeadType = ShapeNodeType
        self.Refresh()
        self.ScrollToCoords((0,0))

        return False

    def Reset(self):
        #dlg = wx.MessageDialog(self, "Are you sure you want to reset your system to the\nunscaled and rotated shape definition?", caption="Confirm reset", style=wx.OK|wx.CANCEL)
        #if not dlg.ShowModal() == wx.ID_OK:
        #    return
        #dlg.Destroy()
        self.assembly.ResetToShapeDef()
        self.InitializeDrawing()
        self.Refresh()

    def Reload(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reload your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", defaultValue=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            SimUtils.globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()

        self.assembly = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        if not SimUtils.globalApp.seedDefinitionFile is None:
            self.assembly.LoadOritatamiAssembly(SimUtils.globalApp.seedDefinitionFile, True)
        elif not SimUtils.globalApp.systemDefinitionFile is None:
            self.assembly.LoadOritatamiSystem(SimUtils.globalApp.systemDefinitionFile, SimUtils.globalApp.systemDefinitionFileFullPath)  
        self.InitializeDrawing()
        self.Refresh()
        
    def Compile(self):

        # Check for connectivity
        self.app.OutputMessage("Checking connectivity of shape...")

        orderedList = self.assembly.GetOrderedList()
        print("Found " + str(len(orderedList)) + " points: " + str(orderedList))
        self.assembly.orderedList = orderedList
        self.assembly.scaledAndRotatedList = []

        if len(orderedList) < self.assembly.numTiles:
            self.app.OutputMessage("The shape is not connected!  Aborting.")
            return False
        self.app.OutputMessage("The shape is connected.  Continuing...")
        
        # Rotate and scale
        topLeft = orderedList[0]
        xb = topLeft[0]
        yb = topLeft[1]
        scaledAndRotated = [topLeft]

        for p in orderedList[1:]:
            x = p[0]
            y = p[1]
            
            xDiff = x - xb
            yDiff = y - yb
            xs = (2 * xDiff) + yDiff
            ys = yDiff - xDiff
            
            newPt = (xb + xs, yb + ys)
            #print(str(p) + " -> " + str(newPt))
            scaledAndRotated.append(newPt)
        
        savedShapePoints = self.assembly.shapePoints
        self.assembly = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        self.assembly.shapePoints = savedShapePoints
        self.assembly.orderedList = orderedList
        self.assembly.scaledAndRotatedList = scaledAndRotated
        self.assembly.gadgetSet = GadgetSet()
        self.assembly.currentGadget = 0

        for p in scaledAndRotated:
            self.assembly.seed.append([ScaledNodeType.name, p])
            loc = ONode("scaledNode", p)
            self.assembly.SetLoc(p, loc)

        #self.assembly.ResetToShapeDef()
        self.app.OutputMessage("Shape has been successfully scaled and rotated.")
        self.InitializeDrawing()
        self.Refresh()
        
    def GenerateOritatamiSystem(self):

        # Generate the Hamiltonian path
        allGadgets = self.assembly.GetAllGadgets()
        self.app.OutputMessage("There are " + str(len(allGadgets)) + " gadgets: " + str(allGadgets))
        
        allNodes = []
        for gCoords in allGadgets:
            allNodes.append(gCoords)
            for n in ONBRS:
                coords = (gCoords[0]+n[0], gCoords[1]+n[1])
                if not coords in allNodes:
                    allNodes.append(coords)
        self.app.OutputMessage("There are " + str(len(allNodes)) + " nodes: " + str(allNodes))
        
        visitedNodes = []
        currentNode = self.assembly.scaledAndRotatedList[0]
        firstNode = currentNode
        bFirst = True
        
        while bFirst == True or not currentNode == firstNode:
            bFirst = False
            #print("currentNode: " + str(currentNode))
            visitedNodes.append(currentNode)

            nbrGadgets = []
            for n in ONBRS:
                pn = (currentNode[0]+n[0], currentNode[1]+n[1])
                g = self.assembly.gadgetSet.Get(pn)
                if not g is None:
                    nbrGadgets.append(g)

            nextNode = None
            for gadget in nbrGadgets:
                for e in gadget.edges:
                    loc1 = (gadget.coords[0]+e[0][0], gadget.coords[1]+e[0][1])
                    loc2 = (gadget.coords[0]+e[1][0], gadget.coords[1]+e[1][1])
                    if loc2 == currentNode and not loc1 in visitedNodes:
                        nextNode = loc1
                        break
                    elif loc1 == currentNode and not loc2 in visitedNodes:
                        nextNode = loc2
                        break
            if nextNode == None:
                #print("All done!")
                break
            currentNode = nextNode
        
        numVis = len(visitedNodes)
        if numVis == len(allNodes):
            self.app.OutputMessage("The Hamiltonian path hit all " + str(numVis) + " nodes: " + str(visitedNodes))
        else:
            self.app.OutputMessage("ERROR: The attempted Hamiltonian path hit only " + str(numVis) + " out of " + str(len(allNodes)) + " nodes: " + str(visitedNodes))

        # Generate the Oritatami system
        gadgetSequence = []
        i = 0
        
        while i < numVis:
            # Get next set of nodes
            n1 = visitedNodes[i]
            n2 = visitedNodes[(i+1)%numVis] # assumes visitedNodes to be a cycle
            n3 = visitedNodes[(i+2)%numVis]

            # Get offsets from middle then compare distance
            diff1 = (n2[0]-n1[0], n2[1]-n1[1])
            diff3 = (n2[0]-n3[0], n2[1]-n3[1])
            nbr1 = ONBRINDEX[diff1]
            nbr3 = ONBRINDEX[diff3]
            angle = (nbr3 - nbr1) % 6
            #print("Angle " + str(n1) + "->" + str(n2) + "->" + str(n3) + " is: " + str(angle))
            gadgetSequence.append(angle)

            i = i + 1
            
        # Get desired output file name from user
        saveFileDialog = wx.FileDialog(self, "Save Oritatami System As", "", "", 
                                        "Oritatami assembly files (*.oass)|*.oass",
                                        wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if not saveFileDialog.ShowModal() == wx.ID_OK:
            SimUtils.globalApp.ShowErrorDialog("You must provide a file name to save the Oritatami system to.")
            return False
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        
        if (not len(fileName) > len(".oass")) or (not fileName[-5:] == ".oass"):
            fileName = fileName + ".oass"
        #if os.path.isabs(fileName):
        # os.path.realpath(__file__)  will get path to executing script
        #    fileName = os.path.relpath(fileName)
        #    print("basename: " + str(os.path.basename(fileName)))
        oassFullFileName = fileName
        oritFullFileName = fileName[:-5] + ".orit"
        oritFileName = os.path.basename(oritFullFileName)
        
        #print("Going to save to " + oassFullFileName + " and " + oritFullFileName + " orit base: " + oritFileName)
        
        # Load the definition files to use as a template

        os.stat(DEFINITION_FILE_OASS)
        rFile = open(DEFINITION_FILE_OASS, "r")
        print("Using Oritatami system definition template '" + DEFINITION_FILE_OASS + "'")
        lines = rFile.readlines()
        rFile.close()
        seedLine = ""
        lastBeadLine = ""
        systemDefLine = "OritatamiSystemDef=" + oritFileName + "\n"
        for line in lines:
            if line.startswith("Assembly="):
                seedLine = line
            elif line.startswith("LastBead="):
                lastBeadLine = line
        
        os.stat(DEFINITION_FILE_ORIT)
        rFile = open(DEFINITION_FILE_ORIT, "r")
        print("Using Oritatami seed definition template '" + DEFINITION_FILE_ORIT + "'")
        lines = rFile.readlines()
        rFile.close()
        beadTypeLine = ""
        delayLengthLine = ""
        arityLine = ""
        lwbLine = ""
        bffsLine = ""
        for line in lines:
            if line.startswith("BeadTypes="):
                beadTypeLine = line
            elif line.startswith("DelayLength="):
                delayLengthLine = line
            elif line.startswith("Arity="):
                arityLine = line
            elif line.startswith("LockWithoutBinding="):
                lwbLine = line
            elif line.startswith("BindFullFloppySegment="):
                bffsLine = line
        beadCycleLine = "BeadCycle=["

        os.stat(DEFINITION_FILE_TXT)
        rFile = open(DEFINITION_FILE_TXT, "r")
        print("Using Oritatami gadget definition template '" + DEFINITION_FILE_TXT + "'")
        lines = rFile.readlines()
        rFile.close()
        gadgetDict = {}
        for line in lines:
            l = line.split("=")
            key = l[0]
            seq = l[1].strip()
            gadgetDict[key] = seq
        
        bFirst = True
        last = None
        
        # the first table below represent colors that are even (b) or odd (g) steps away from a red 'r' tile with previous tile blue 'b'
        # other table entries are similar 
        colorTable      = {}
        colorTable[('r','b')] = ['b','g']
        colorTable[('r','g')] = ['g','b']
        colorTable[('g','b')] = ['b','r']
        colorTable[('g','r')] = ['r','b']
        colorTable[('b','g')] = ['g','r']
        colorTable[('b','r')] = ['r','g']
        
        lastColor = None
        lastLastColor = None
        
        
        # TODO: seed to get the correct rotation of the shape. The following definition hard codes the seed. 
        seedLine=Assembly=[['1r', (0, 0), None, []], ['2r', (1, 0), (0, 0), [(1, 1)]], ['3r', (2, 0), (1, 0), []], ['4r', (2, -1), (2, 0), [(1, 0)]]]
        
        
        
        for g in gadgetSequence:
            if bFirst == False:
                beadCycleLine = beadCycleLine + ","
                
            #print("Appending gadget " + str(g))
            if bFirst == True:
                
                # create the 4 bead seed line
                firstGadget = gadgetDict[str(g) + 'r'].split(",")
                for i in range(4):
                    seedLine[i][0] = firstGadget[i].replace("'","")
                seedLine = "Assembly=" + str(seedLine) + "\n"

                # The first gadget in the transcript is "missing" 4 beads, which are part of the seed
                beadCycleLine = beadCycleLine + ",".join(firstGadget[4:])
                
                
                lastColor = 'r'
                lastLastColor = 'g'
                bFirst = False
            else:                        
                currentColor = colorTable[(lastColor,lastLastColor)][last % 2]
                beadCycleLine = beadCycleLine + gadgetDict[str(g) + currentColor]
                lastLastColor = lastColor
                lastColor = currentColor
                
            last = g
            
        #print("Appending final gadget " + str(1))
        # beadCycleLine = beadCycleLine + "," + gadgetDict[str(1) + colorTable[(lastColor,lastLastColor)][last % 2]]

        beadCycleLine = beadCycleLine + "]\n"
        
        #print len(ast.literal_eval(beadCycleLine.split("=")[1]))
        #print("beadCycleLine: " + beadCycleLine)
        

        wFile = open(oassFullFileName, "w")
        wFile.write(systemDefLine)
        wFile.write(seedLine)
        wFile.write(lastBeadLine)
        wFile.close()
        
        wFile = open(oritFullFileName, "w")
        wFile.write(beadTypeLine)
        wFile.write(beadCycleLine)
        wFile.write(delayLengthLine)
        wFile.write(arityLine)
        wFile.write(lwbLine)
        wFile.write(bffsLine)
        wFile.close()

        self.app.OutputMessage("Saved to output files " + oritFullFileName + " and " + oassFullFileName)

    def AddGadget(self):

        if self.assembly.scaledAndRotatedList is None or len(self.assembly.scaledAndRotatedList) == 0:
            self.app.OutputMessage("Unable to add a gadget.  A connected shape must first be scaled and rotated.")
            return False

        if self.assembly.currentGadget >= len(self.assembly.scaledAndRotatedList):
            self.app.OutputMessage("There are no more points to replace with gadgets.")
            return False
        
        # Replace next scaled point with a gadget
        p = self.assembly.scaledAndRotatedList[self.assembly.currentGadget]
        self.assembly.currentGadget = self.assembly.currentGadget + 1
        ng = NodeGadget(p)
        self.assembly.gadgetSet.Add(ng)

        abstractNbrPattern = self.assembly.gadgetSet.GetAbstractNbrPattern(p)
        #print("Got abstractNbrPattern " + str(abstractNbrPattern))
        (matchedPattern, rot, reflect, changeList) = MatchPattern(abstractNbrPattern)
        
        if not matchedPattern is None:
            if self.assembly.gadgetSet.ExtendCycle(ng.coords, rot, reflect, changeList) == True:
                self.InitializeDrawing()
                self.Refresh()
                return True
        else:
            self.app.OutputMessage("ERROR:  Could not find matching pattern for neighboring gadgets!")
            return False


    def DoForwardStep(self):

        return self.AddGadget()
        
        #MJP TODO: Switch back to efficient drawing later.  For now, just redraw everything for each step!
        #self.InitializeDrawing(self.pdc)
        #self.ScrollToCoords(coords)

    def DoGo(self):
        bContinue = self.DoForwardStep()
        while bContinue == True:
            bContinue = self.DoForwardStep()

    def ZoomThenRecenter(self, zoomIn=True):
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
        coords = self.GetCoordsFromPixels(center[0], center[1], True)
        self.DoScale(zoomIn, scrollToCoords=coords)
    
    def ConvertEventCoords(self, event):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (event.GetX() + (xView * xDelta),
            event.GetY() + (yView * yDelta))

    def OffsetRect(self, r):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        r.OffsetXY(-(xView*xDelta),-(yView*yDelta))

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("a") or keycode == ord("A"):
            #self.AddTileCopiesToDrawing()
            #self.Refresh()
            return
        elif keycode == ord("p") or keycode == ord("P"):
            self.PrintDebugInfo()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset()
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSaveAll()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad(bAssembly=True, bSystem=True)
        elif keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("=") or keycode == ord("+"):
            self.DoScale(True, bScrollToCenter=True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.DoScale(False, bScrollToCenter=True)
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("m") or keycode == ord("M"):
            self.MessageToggle(not self.app.showMessages)
            self.app.menuView.Check(self.app.itemTM.GetId(), self.app.showMessages)
        elif keycode == ord("d") or keycode == ord("D"):
            self.DisplayToggle(not self.app.showDisplayUpdates)
            self.app.menuView.Check(self.app.itemTD.GetId(), self.app.showDisplayUpdates)
        elif keycode == ord("b") or keycode == ord("B"):
            self.SetBreakpoints()
        elif keycode == wx.WXK_F2:
            self.Refresh()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def ShowHelp(self):
        self.app.OnHelp(None)

    def PrintDebugInfo(self):        
        #self.app.OutputMessage("some stuff")
        return

    def OnMouse(self, event):
        
        #=======================================================================
        # if event.IsPageScroll():
        #     print("is page scroll")
        # if event.MiddleIsDown():
        #     print("middle is down")
        #     if event.Dragging():
        #         print("Dragging")
        #=======================================================================
            
        x,y = self.ConvertEventCoords(event)
        
        if self.showMouseOver == True:
            x2,y2 = self.GetCoordsFromPixels(x, y)
    
            if x2 == None:
                self.frame.SetStatusText("", 0)
                self.frame.SetStatusText("", 1)
            else:
                self.frame.SetStatusText("Coords: (" + str(int(x2)) + "," + str(int(y2)) + ")", 0)
                onode = self.assembly.GetLoc((x2,y2))
                if not onode is None:
                    outStr = "Bead type: " + str(onode.type)
                    self.frame.SetStatusText(outStr, 1)
                else:
                    self.frame.SetStatusText("", 1)
        
        if event.LeftDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if self.drawingModeBeadType is None:
                    SimUtils.globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                else:
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.assembly.GetLoc((x2, y2)) is None:
                            parentLoc = None
                            if (self.assembly.lastBeadLoc is not None) and self.assembly.AreNbrLocs((x2, y2), self.assembly.lastBeadLoc) == True:
                                parentLoc = self.assembly.lastBeadLoc
                            self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.negSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.assembly.GetLoc(self.linkStart)
                        if self.negSelected is not None:
                            saved = self.negSelected.coords
                            self.negSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.posSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.assembly.GetLoc(self.linkStart)
                        self.posSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.assembly.GetLoc(self.linkStart)
                            childNode = self.assembly.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (self.assembly.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        if parentNode.child is not None:
                                            SimUtils.globalApp.ShowErrorDialog("The node at " + str(self.linkStart) + " already has a child")
                                        elif childNode.parent is not None:
                                            SimUtils.globalApp.ShowErrorDialog("The node at " + str(self.linkEnd) + " already has a parent")
                                        else:
                                            bAlreadyBonded = False
                                            for n in range(parentNode.NumBonds()):
                                                b = parentNode.GetNthBond(n)
                                                if b == self.linkEnd:
                                                    bAlreadyBonded = True
                                            if bAlreadyBonded == True:
                                                SimUtils.globalApp.ShowErrorDialog("Nodes must not already be bonded to be linked as parent and child")
                                            else:
                                                parentNode.child = childNode
                                                childNode.parent = parentNode
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Nodes must be neighbors to be linked")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    if (self.assembly.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        bAlreadyBonded = False
                                        for n in range(parentNode.NumBonds()):
                                            b = parentNode.GetNthBond(n)
                                            if b == self.linkEnd:
                                                bAlreadyBonded = True
                                        if bAlreadyBonded == True:
                                            SimUtils.globalApp.ShowErrorDialog("Nodes are already bonded")
                                        elif (parentNode.child == childNode) or (childNode.child == parentNode):
                                            SimUtils.globalApp.ShowErrorDialog("Nodes cannot be bonded if they are linked as parent and child")
                                        else:
                                            bDoIt = True
                                            numP = parentNode.NumBonds()
                                            if (self.confirmOverArity == True) and (numP >= self.assembly.arity):
                                                msg = "The bead at location " + str(parentNode.coords) + " currently has " + str(numP)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.assembly.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SIMU_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == SIMU_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            numC = childNode.NumBonds()
                                            if (bDoIt == True) and (self.confirmOverArity == True) and (numC >= self.assembly.arity):
                                                msg = "The bead at location " + str(childNode.coords) + " currently has " + str(numC)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.assembly.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SIMU_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == SIMU_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            if (bDoIt == True) and (self.confirmInvalidBonds == True) and (self.assembly.ONodeTypes[parentNode.type].BondsToNodeType(childNode.type) == False):
                                                msg = "Bead types '" + str(parentNode.type) + "' and '" + str(childNode.type) + "' don't form bonds."
                                                msg = msg + "\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm bond between non-bonding types", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == SIMU_ID_OK_NOCHECK:
                                                    self.confirmInvalidBonds = True
                                                elif retId == SIMU_ID_OK_CHECK:
                                                    self.confirmInvalidBonds = False
                                                else:
                                                    bDoIt = False
                                            if bDoIt == True:
                                                self.assembly.AddBond(self.linkStart, self.linkEnd)
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Nodes must be neighbors to be bonded")

        elif event.RightDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                #self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                doScale = True
                if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                    doScale = False
                x2,y2 = self.GetCoordsFromPixels(x, y, True)
                self.DoScale(doScale, scrollToCoords=(x2,y2))
                sz = self.GetSize()
                self.WarpPointer(sz[0]/2, sz[1]/2)
                #self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                x2,y2 = self.GetCoordsFromPixels(x, y)
                if (x2 is not None) and (y2 is not None):
                    if self.assembly.GetLoc((x2, y2)) is not None:
                        self.ManuallyRemoveNode((x2, y2))
                        
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.posSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.assembly.GetLoc(self.linkStart)
                        if self.posSelected is not None:
                            saved = self.posSelected.coords
                            self.posSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.negSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.assembly.GetLoc(self.linkStart)
                        self.negSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.assembly.GetLoc(self.linkStart)
                            childNode = self.assembly.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                bDidIt = False
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (parentNode.child == childNode):
                                        parentNode.child = None
                                        childNode.parent = None
                                        bDidIt = True
                                    elif (childNode.child == parentNode):
                                        childNode.child = None
                                        parentNode.parent = None
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing()
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Cannot remove link between beads which aren't linked.")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    bAlreadyBonded = False
                                    for n in range(parentNode.NumBonds()):
                                        b = parentNode.GetNthBond(n)
                                        if b == self.linkEnd:
                                            bAlreadyBonded = True
                                    if bAlreadyBonded == True:
                                        self.assembly.RemoveBond(self.linkStart, self.linkEnd)
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing()
                                    else:
                                        SimUtils.globalApp.ShowErrorDialog("Cannot remove bond between beads which don't have a bond.")

        if event.LeftUp():
            self.dragStart = None
            self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))

        elif event.Dragging():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                if self.dragStart is not None:
                    #self.SetScrollRate(1,1)
                    new = self.ScreenToClient(wx.GetMousePosition())
                    dx = int((self.dragStart.x-new.x)/1)
                    dy = int((self.dragStart.y-new.y)/1)
                    spx = self.GetScrollPos(wx.HORIZONTAL)
                    spy = self.GetScrollPos(wx.VERTICAL)
                    self.Scroll(spx+dx, spy+dy)
                    self.dragStart = new
            
            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if event.LeftIsDown():
                    if self.drawingModeBeadType is None:
                        SimUtils.globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                    else:
                        x2,y2 = self.GetCoordsFromPixels(x, y)
                        if (x2 is not None) and (y2 is not None):
                            if self.assembly.GetLoc((x2, y2)) is None:
                                parentLoc = None
                                if (self.assembly.lastBeadLoc is not None) and self.assembly.AreNbrLocs((x2, y2), self.assembly.lastBeadLoc) == True:
                                    parentLoc = self.assembly.lastBeadLoc
                                self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
                elif event.RightIsDown():
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.assembly.GetLoc((x2, y2)) is not None:
                            self.ManuallyRemoveNode((x2, y2))

    def ManuallyAddNode(self, typeName, coords, parentCoords):
        self.assembly.AddNode(self.drawingModeBeadType.name, coords, parentCoords)
        self.ManuallyDrawNodes([coords])
        self.frame.SetStatusText("Number of beads: " + str(self.assembly.GetNumBeads()), 2)
        
    def ManuallyRemoveNode(self, coords):
        self.assembly.RemoveNode(coords)
        self.InitializeDrawing()
        self.frame.SetStatusText("Number of beads: " + str(self.assembly.GetNumBeads()), 2)
        
    def ManuallyDrawNodes(self, coordList):
        self._lock.acquire()
        for coords in coordList:
            self.DrawQueue.append(coords)
        self._lock.release()
        self.DrawAssemblyUpdate("new")

    def ScrollToCoords(self, coords):
        if (not coords == None) and (not coords[0] == None) and (not coords[1] == None):
            (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
            
            # Some slight adjustments due to positioning of triangular grid
            x = x - self.sideLen
            y = y - self.sideLen

            x = x * self.drawScale
            y = y * self.drawScale
            (xs, ys) = self.GetScrollPixelsPerUnit()
            (xsz,ysz) = self.GetSize()
            self.Scroll((x-xsz/2+self.sideLen*self.drawScale)/xs, (y-ysz/2+self.sideLen*self.drawScale)/ys)

    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, colorName):
        if not self.brush_cache.has_key(colorName):
            if type(colorName) is str:
                if "(" in colorName:
                    colorTuple = eval(colorName)
                    if len(colorTuple) > 3:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2], colorTuple[3])
                    else:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2])
                else:
                    color = wx.TheColourDatabase.Find(colorName)
                self.brush_cache[colorName] = wx.Brush(color)
        return self.brush_cache[colorName]


#===============================================================================
#     def OnPaint(self, event):
#         # Create a buffered paint DC.  It will create the real
#         # wx.PaintDC and then blit the bitmap to it when dc is
#         # deleted.  
#         dc = wx.BufferedPaintDC(self)
#         # use PrepateDC to set position correctly
#         self.PrepareDC(dc)
#         # we need to clear the dc BEFORE calling PrepareDC
#         bg = wx.Brush(self.GetBackgroundColour())
#         dc.SetBackground(bg)
#         dc.Clear()
# 
#         dc.SetUserScale(self.drawScale, self.drawScale)
# 
#         # create a clipping rect from our position and size
#         # and the Update Region
#         xv, yv = self.GetViewStart()
#         dx, dy = self.GetScrollPixelsPerUnit()
#         x, y   = (xv * dx, yv * dy)
#         rgn = self.GetUpdateRegion()
#         rgn.Offset(x,y)
#         r = rgn.GetBox()
#         # draw to the dc using the calculated clipping rect
#         self.pdc.DrawToDCClipped(dc,r)
#===============================================================================

        
    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

    
        dc = wx.ClientDC(self)
        
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new"):
            
            self._lock.acquire()
            copiedList = list(self.DrawQueue)
            if self.lastCurrent is not None:    # Redraw the bead that was current at the last step
                copiedList.append(self.lastCurrent)
            self.DrawQueue = []
            self._lock.release()
            
            for e in copiedList:
                onode = self.assembly.GetLoc((e[0], e[1]))
                if onode is not None:
                    self.DrawAtCoords(dc, e[0], e[1], onode, self.assembly)
    
            if len(self.RemoveQueue) > 0:
                self.RemoveQueue = []
                self.DoRedrawAssembly()

        
    def DrawGrid(self, dc):
        
        # Compute new dimensions
        self.gridWidth = self.assembly.GetWidth()
        self.gridHeight = self.assembly.GetHeight()
        
        self.winWidth  = self.sideLen * self.gridWidth  + self.GetHorizOffset(self.gridWidth)
        self.winHeight = self.GetVertOffset(self.gridHeight)

        self.SetVirtualSize((self.winWidth*self.drawScale, self.winHeight*self.drawScale))

        #  Remove old data
        dc.Clear()

        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                onode = self.assembly.GetLoc((i,j))
                if onode is not None:
                    if onode.type == "scaledNode":
                        gs = self.assembly.gadgetSet.Get(onode.coords)
                        if not gs is None:
                            gs.DrawPolygon(dc, self)

        dc.SetPen(self.gridPen)
        if (self.showGrid == SHOW_GRID_AND_COORDS) or (self.showGrid == SHOW_GRID_ONLY):
            # Draw horizontal lines            
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(self.assembly.minX, i)
                endPt = self.GetPixelsForCoords(self.assembly.maxX, i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

            # Draw positive sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.assembly.minY)
                endPt = self.GetPixelsForCoords(i, self.assembly.maxY)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                startPt = self.GetPixelsForCoords(self.assembly.maxX, i)
                endPt = self.GetPixelsForCoords(i, self.assembly.maxY)
                #dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

            # Draw negative sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.assembly.minY)
                endPt = self.GetPixelsForCoords(self.assembly.minX,i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                startPt = self.GetPixelsForCoords(i, self.assembly.maxY)
                endPt = self.GetPixelsForCoords(self.assembly.maxX, i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

        if self.showGrid == SHOW_GRID_AND_COORDS:
            for i in range(self.assembly.minX, self.assembly.maxX+1):
                for j in range(self.assembly.minY, self.assembly.maxY+1):
                    (xp,yp) = self.GetPixelsForCoords(i,j)
                    dc.DrawText(str(i) + "," + str(j), xp+8, yp-2)

        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                onode = self.assembly.GetLoc((i,j))
                if onode is not None:
                    if onode.type == "shapeNode":
                        self.DrawAtCoords(dc, i, j, onode, self.assembly)
                    elif onode.type == "scaledNode":
                        gs = self.assembly.gadgetSet.Get(onode.coords)
                        if not gs is None:
                            gs.Draw(dc, self)
                        else:
                            self.DrawAtCoords(dc, i, j, onode, self.assembly)

        #self.frame.SetStatusText("Number of beads: " + str(self.assembly.GetNumBeads()), 2)


    def GetHorizOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(60))
    
    def GetVertOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(30))

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoordAdj = xCoord - self.assembly.minX
        yCoordAdj = yCoord - self.assembly.minY
        xPixel = (xCoordAdj * self.sideLen) + self.GetHorizOffset(yCoordAdj)
        yPixel = self.GetVertOffset(yCoordAdj)        
        yPixel = self.winHeight - (yPixel + self.sideLen/2)
        return (xPixel, yPixel)

    def GetCoordsFromPixels(self, xPixel, yPixel, bGapOkay=False):
        sz = self.GetVirtualSize()
        yPixel = sz[1] - (yPixel + (self.sideLen/2)*self.drawScale)
        xPixel = xPixel / self.drawScale
        yPixel = yPixel / self.drawScale
        yCoord = (yPixel / (self.sideLen * math.cos(math.radians(30)))) + self.assembly.minX
        xCoord = ((xPixel - self.GetHorizOffset(yCoord - self.assembly.minY)) / self.sideLen) + self.assembly.minX
        (x2,x1) = math.modf(xCoord)
        (y2,y1) = math.modf(yCoord)
        if ((abs(x2) < 0.2 or abs(x2) > 0.8) and (abs(y2) < 0.2 or abs(y2) > 0.8)) or (bGapOkay == True):
            retX = int(round(xCoord))
            retY = int(round(yCoord))
            if (retX >= self.assembly.minX and retX <= self.assembly.maxX and retY >= self.assembly.minY and retY <= self.assembly.maxY):
                return (retX, retY)
        return (None,None)

    def DrawAtCoords(self, dc, xCoord, yCoord, onode, grid):
        
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)

        dc.SetPen(self.gridPen)

        if onode.type in grid.ONodeTypes:
            nodeColor = grid.ONodeTypes[onode.type].color
        else:
            nodeColor = "Black"
        pointBrush = self.ColoredBrush(nodeColor)
        dc.SetBrush(pointBrush)
        dc.DrawCircle((xPixel,yPixel), 10)

        if onode == self.posSelected:
            pointBrush = self.ColoredBrush("Green")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode == self.negSelected:
            pointBrush = self.ColoredBrush("Red")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode.coords == self.assembly.lastBeadLoc:
            #pointBrush = self.ColoredBrush("White")
            #dc.SetBrush(pointBrush)
            #dc.DrawCirclePoint((xPixel,yPixel), 5)
            self.lastCurrent = onode.coords
        
    def ToggleGrid(self):
            self.showGrid = (self.showGrid + 1) % 3
            self.InitializeDrawing()
            
    def MessageToggle(self, bShow):
        self.app.showMessages = bShow
        if bShow == True:
            for msg in self.app.messageQueue:
                self.app.OutputMessage(msg)
        self.app.messageQueue = []

    def DisplayToggle(self, bShow):
        self.app.showDisplayUpdates = bShow
        if self.app.showDisplayUpdates == True:
            self.Refresh()

